from ..utils import convert_to_snake_case

PRINT_IMPORTS = ['from django.forms.models import model_to_dict',
                 'from django.conf import settings',
                 'from django.test import TestCase',
                 'from django.db import IntegrityError'
                 ]
PRINT_TEST_CLASS = """
class TestCase{0}(TestCase):

    def test_create(self):
        \"\"\"
        Test the creation of a {0} model using a factory
        \"\"\"
        {1} = {0}Factory.create()
        self.assertEqual({0}.objects.count(), 1)

    def test_create_batch(self):
        \"\"\"
        Test the creation of 5 {0} models using a factory
        \"\"\"
        {1}s = {0}Factory.create_batch(5)
        self.assertEqual({0}.objects.count(), 5)
        self.assertEqual(len({1}s), 5)
"""

PRINT_TEST_ATTRIBUTE_COUNT = """
    def test_attribute_count(self):
        \"\"\"
        Test that all attributes of {0} server are counted. It will count the primary key and all editable attributes.
        This test should break if a new attribute is added.
        \"\"\"
        {1} = {0}Factory.create()
        {1}_dict = model_to_dict({1})
        self.assertEqual(len({1}_dict.keys()), {2})

"""

PRINT_TEST_ATTRIBUTE_CONTENT = """
    def test_attribute_content(self):
        \"\"\"
        Test that all attributes of {0} server have content. This test will break if an attributes name is changed.
        \"\"\"
        {1} = {0}Factory.create()
"""

PRINT_TEST_ATTRIBUTE_UNIQUE = """
    def test_{2}_is_unique(self):
        \"\"\"
        Tests attribute {2} of model {0} to see if the unique constraint works.
        This test should break if the unique attribute is changed.
        \"\"\"
        {1} = {0}Factory.create()
        {1}_02 = {0}Factory.create()
        {1}_02.{2} = {1}.{2}
        try:
            {1}_02.save()
            self.fail('Test should have raised and integrity error')
        except IntegrityError as e:
            self.assertEqual(str(e), '') #FIXME This test is incomplete
"""


# noinspection PyProtectedMember,PyProtectedMember
class ModelTestCaseGenerator(object):
    def __init__(self, model):
        self.model = model

    def _generate(self):
        model_test_case_content = list()
        model_test_case_content.append({'print': PRINT_TEST_CLASS,
                                        'args': [self.model.__name__,
                                                 convert_to_snake_case(self.model.__name__)]})
        model_test_case_content.append({'print': PRINT_TEST_ATTRIBUTE_COUNT,
                                        'args': [self.model.__name__,
                                                 convert_to_snake_case(self.model.__name__),
                                                 len(self.model._meta.fields)]})

        content_text = PRINT_TEST_ATTRIBUTE_CONTENT.format(self.model.__name__,
                                                           convert_to_snake_case(self.model.__name__))
        PRINT_IMPORTS.append('from {} import {}'.format(self.model.__module__, self.model.__name__))
        PRINT_IMPORTS.append('from ..factories import {}Factory'.format(self.model.__name__))

        for field in self.model._meta.fields:
            field_type = type(field).__name__
            field_data = dict()
            assertion = '        self.assertIsNotNone({0}.{1})\n'.format(convert_to_snake_case(self.model.__name__),
                                                                         field.name)
            content_text += assertion
        model_test_case_content.append({'print': content_text, 'args': None})
        # Build unique tests
        add_integrity_error_to_imports = False
        for field in self.model._meta.fields:
            if field.unique and not field.primary_key:
                data = [self.model.__name__,
                        convert_to_snake_case(self.model.__name__),
                        field.name]
                unique_test = PRINT_TEST_ATTRIBUTE_UNIQUE.format(*data)
                model_test_case_content.append({'print': unique_test, 'args': None})
                add_integrity_error_to_imports = True
        if add_integrity_error_to_imports:
            PRINT_IMPORTS.append('from django.db import IntegrityError')

        return model_test_case_content

    def __str__(self):
        printable = list()
        for print_data in self._generate():
            try:
                if print_data['args'] is not None:
                    printable.append(print_data['print'].format(*print_data['args']))
                else:
                    printable.append(print_data['print'])
            except IndexError as e:
                print('-' * 74)
                print('{print} {args}'.format(**print_data))
                raise e

        return '\n'.join(printable)


class AppModelsTestCaseGenerator(object):
    def __init__(self, app):
        self.app = app

    def _generate(self):
        app_content = list()
        for model in self.app.get_models():
            model_test_case = ModelTestCaseGenerator(model)
            app_content.append(model_test_case)
        return app_content

    def _get_imports(self):
        imports_to_print = list(set(PRINT_IMPORTS))
        imports_to_print.sort(reverse=True)
        return '\n'.join(imports_to_print)

    def __str__(self):
        printable = list()
        printable.append(self._get_imports())
        for model_test_cases in self._generate():
            printable.append(str(model_test_cases))

        return '\n'.join(printable)



# -*- coding: utf-8
from django.apps import AppConfig




class DjangoTestToolsConfig(AppConfig):
    name = 'django_test_tools'

    def ready(self):
        from django_test_tools.checks import example_check




import logging

logger = logging.getLogger(__name__)
PRINT_IMPORTS = ['from rest_framework import serializers', ]
PRINT_SERIALIZER_CLASS = """
class {0}Serializer(serializers.{1}):
    \"\"\"
    Standard Serializer for {0} model.
    \"\"\"

    class Meta:
        model = {0}
        fields = ({2})
"""


class SerializerGenerator(object):
    def __init__(self, model, serializer_class='ModelSerializer'):
        self.model = model
        self.serializer_class = serializer_class

    def _generate(self):
        model_test_case_content = list()

        field_names = list()
        # noinspection PyProtectedMember
        for field in self.model._meta.fields:
            field_names.append('\'{}\''.format(field.name))

        model_test_case_content.append({'print': PRINT_SERIALIZER_CLASS,
                                        'args': [self.model.__name__,
                                                 self.serializer_class,
                                                 ', '.join(field_names)]})
        return model_test_case_content

    def __str__(self):
        printable = list()
        for print_data in self._generate():
            try:
                if print_data['args'] is not None:
                    printable.append(print_data['print'].format(*print_data['args']))
                else:
                    printable.append(print_data['print'])
            except IndexError as e:
                print('-' * 74)
                print('{print} {args}'.format(**print_data))
                raise e

        return '\n'.join(printable)


class AppSerializerGenerator(object):
    def __init__(self, app, serializer_class='ModelSerializer'):
        self.app = app
        self.serializer_class = serializer_class

    def _generate(self):
        app_content = list()
        for model in self.app.get_models():
            PRINT_IMPORTS.append('from {} import {}'.format(model.__module__, model.__name__))
            logger.debug('IMPORTS: {}'.format(PRINT_IMPORTS))
            model_test_case = SerializerGenerator(model, self.serializer_class)
            app_content.append(model_test_case)
        return app_content

    def _get_imports(self):
        logger.debug('IMPORTS: {}'.format(PRINT_IMPORTS))
        imports_to_print = list(set(PRINT_IMPORTS))
        imports_to_print.sort(reverse=True)
        return '\n'.join(imports_to_print)

    def __str__(self):
        printable = list()
        serializers_codes = self._generate()
        printable.append(self._get_imports())
        for serializers_code in serializers_codes:
            printable.append(str(serializers_code))

        return '\n'.join(printable)



import os

from django.template.loader import render_to_string

from ..app_manager import DjangoAppManager


def create_folder_structure(doc_base_folder, project_name):
    # print('**** APP_DIR {} ****'.format(settings.APPS_DIR))
    app_manager = DjangoAppManager()
    project_apps = app_manager.get_project_apps(project_name)
    project_folder = os.path.join(doc_base_folder, project_name)
    for app_name, app in project_apps.items():
        folder = os.path.join(project_folder, app.label)
        if not os.path.exists(folder):
            os.makedirs(folder)
        data = {
            'verbose_name': str(app.verbose_name),
            'app_package': app_name
        }
        template = 'django_test_tools/app_index.rst.j2'
        write_template(data, folder, 'index.rst', template)
        data = {
            'verbose_name': str(app.verbose_name),
            'module_name': '{}.models'.format(app_name)
        }
        template = 'django_test_tools/app_module.rst.j2'
        write_template(data, folder, '{}.models.rst'.format(app_name), template)


def get_module_files(folder):
    file_black_list = ['__init__.py', 'urls.py']
    folder_black_list = ['tests', 'migrations']
    file_list = list()
    package_name = ''
    for root, dirs, files in os.walk(folder):
        base_folder_name = os.path.split(root)[1]
        if base_folder_name not in folder_black_list:
            if '__init__.py' in files:
                package_name = package_name + '.' + base_folder_name
            for file in files:
                if file.endswith('.py'):
                    module_name = file.split('.')[0]
                    if file not in file_black_list:
                        module_dict = dict()
                        module_dict['filename'] = os.path.join(root, file)
                        module_dict['package_name'] = (package_name + '.{}'.format(module_name))[1:]
                        file_list.append(module_dict)
        else:
            del dirs[:]
    return file_list


def write_template(data, folder, output_file, template):
    rendered = render_to_string(template, data)
    index_filename = os.path.join(folder, output_file)
    with open(index_filename, 'w', encoding='utf-8') as index_file:
        index_file.write(rendered)



import os
import subprocess
from tempfile import NamedTemporaryFile


class GenericCVS(object):
    @classmethod
    def commit(cls, message):
        f = NamedTemporaryFile('wb', delete=False)
        f.write(message.encode('utf-8'))
        f.close()
        subprocess.check_output(cls._COMMIT_COMMAND + [f.name], env=dict(
            list(os.environ.items()) + [(b'HGENCODING', b'utf-8')]
        ))
        os.unlink(f.name)

    @classmethod
    def is_usable(cls):
        try:
            return subprocess.call(
                cls._TEST_USABLE_COMMAND,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE
            ) == 0
        except OSError as e:
            if e.errno == 2:
                # mercurial is not installed then, ok.
                return False
            raise


class Git(GenericCVS):
    """
    Option	Description of Output
    %H      Commit hash
    %h      Abbreviated commit hash
    %T      Tree hash
    %t      Abbreviated tree hash
    %P      Parent hashes
    %p      Abbreviated parent hashes
    %an     Author name
    %ae     Author email
    %ad     Author date (format respects the --date=option)
    %ar     Author date, relative
    %cn     Committer name
    %ce     Committer email
    %cd     Committer date
    %cr     Committer date, relative
    %s      Subject
    """

    _TEST_USABLE_COMMAND = ["git", "rev-parse", "--git-dir"]
    _COMMIT_COMMAND = ["git", "commit", "-F"]

    def __init__(self):
        self.fields = ['%H', '%an', '%ae', '%ad', '%s']

    @classmethod
    def assert_nondirty(cls):
        lines = [
            line.strip() for line in
            subprocess.check_output(
                ["git", "status", "--porcelain"]).splitlines()
            if not line.strip().startswith(b"??")
        ]

        if lines:
            raise Exception(
                "Git working directory is not clean:\n{}".format(
                    b"\n".join(lines)))

    def report(self):
        # git log --pretty=format:"%h - %an, %ad : %s" --date=iso
        #
        # git log --pretty="%h - %s" --author=gitster --since="2008-10-01"
        # --before="2008-11-01" --no-merges -- t/
        format = '|'.join(self.fields)
        try:
            # git-describe doesn't update the git-index, so we do that
            subprocess.check_output(["git", "update-index", "--refresh"])

            # get info about the latest tag in git
            describe_out = subprocess.check_output([
                "git",
                "log",
                '--date=iso',
                '--pretty=format:"{}"'.format(format)
            ], stderr=subprocess.STDOUT
            ).decode()
        except subprocess.CalledProcessError:
            # logger.warn("Error when running git describe")
            return {}

        return describe_out.split('\n')

    @classmethod
    def latest_tag_info(cls):
        try:
            # git-describe doesn't update the git-index, so we do that
            subprocess.check_output(["git", "update-index", "--refresh"])

            # get info about the latest tag in git
            describe_out = subprocess.check_output([
                "git",
                "describe",
                "--dirty",
                "--tags",
                "--long",
                "--abbrev=40",
                "--match=v*",
            ], stderr=subprocess.STDOUT
            ).decode().split("-")
        except subprocess.CalledProcessError:
            # logger.warn("Error when running git describe")
            return {}

        info = {}

        if describe_out[-1].strip() == "dirty":
            info["dirty"] = True
            describe_out.pop()

        info["commit_sha"] = describe_out.pop().lstrip("g")
        info["distance_to_latest_tag"] = int(describe_out.pop())
        info["current_version"] = "-".join(describe_out).lstrip("v")

        return info

    @classmethod
    def add_path(cls, path):
        subprocess.check_output(["git", "add", "--update", path])

    @classmethod
    def tag(cls, name, message):
        command = ["git", "tag", name]
        if message:
            command += ['--message', message]
        subprocess.check_output(command)



import csv
import re


class Flake8Parser(object):
    """
    3     E124 closing bracket does not match visual indentation
    6     E127 continuation line over-indented for visual indent
    11    E128 continuation line under-indented for visual indent
    2     E221 multiple spaces before operator
    1     E222 multiple spaces after operator
    10    E225 missing whitespace around operator
    6     E231 missing whitespace after ','
    2     E251 unexpected spaces around keyword / parameter equals
    4     E261 at least two spaces before inline comment
    4     E262 inline comment should start with '# '
    8     E265 block comment should start with '# '
    4     E266 too many leading '#' for block comment
    2     E271 multiple spaces after keyword
    5     E302 expected 2 blank lines, found 1
    7     E303 too many blank lines (3)
    2     E402 module level import not at top of file
    8     E501 line too long (123 > 120 characters)
    17    F401 'django.contrib.admin' imported but unused
    25    F405 'env' may be undefined, or defined from star imports: .base
    1     F811 redefinition of unused 'RemarksManager' from line 3
    7     F841 local variable 'response' is assigned to but never used
    2     W293 blank line contains whitespace
    6     W391 blank line at end of file
    """

    def __init__(self):
        self.summary_regexp = re.compile(r'(\d+)\s+((?:W|E|F)\d+)\s([\w\s\'.#,()>:]*)')
        self.fieldnames = ['count', 'rule', 'description']

    def parse_summary(self, filename):
        pep8_findings = list()
        with open(filename, 'r', encoding='utf-8') as flake_file:
            for line in flake_file.readlines():
                match = self.summary_regexp.match(line.rstrip('\n'))
                if match:
                    data = dict()
                    data['count'] = match.group(1)
                    data['rule'] = match.group(2)
                    data['description'] = match.group(3)
                    pep8_findings.append(data)
        return pep8_findings

    def write_summary(self, source_filename, target_filename):
        pep8_findings = self.parse_summary(source_filename)
        with open(target_filename, 'w', encoding='utf-8') as csv_file:
            csv_writer = csv.DictWriter(csv_file, fieldnames=self.fieldnames)
            csv_writer.writeheader()
            for data in pep8_findings:
                csv_writer.writerow(data)


class RadonParser(object):
    """


    config/settings/test.py
        LOC: 61
        LLOC: 12
        SLOC: 23
        Comments: 23
        Single comments: 22
        Multi: 4
        Blank: 12
        - Comment Stats
            (C % L): 38%
            (C % S): 100%
            (C + M % L): 44%
    ** Total **
        LOC: 2149
        LLOC: 894
        SLOC: 1311
        Comments: 335
        Single comments: 310
        Multi: 128
        Blank: 400
        - Comment Stats
            (C % L): 16%
            (C % S): 26%
            (C + M % L): 22%
    """

    def __init__(self):
        self.totals_match_regexp = re.compile(r'\s*\*\*\sTotal\s\*\*')
        self.results_regexp = re.compile(r'^\s+(LOC|LLOC|SLOC|Comments|Single\scomments|Multi|Blank):\s(\d+)')
        self.fieldnames = ['loc', 'lloc', 'sloc', 'comments', 'single_comments', 'multi', 'blank']

    def parse_totals(self, filename):
        data = dict()
        totals_found = False
        with open(filename, 'r', encoding='utf-8') as flake_file:
            for line in flake_file.readlines():
                match = self.totals_match_regexp.match(line.rstrip('\n'))
                if match:
                    totals_found = True
                if totals_found:
                    data_match = self.results_regexp.match(line.rstrip('\n'))
                    if data_match:
                        key = data_match.group(1).lower().replace(' ', '_')
                        data[key] = int(data_match.group(2))
        return data

    def write_totals(self, source_filename, target_filename):
        radon_findings = self.parse_totals(source_filename)
        with open(target_filename, 'w', encoding='utf-8') as csv_file:
            csv_writer = csv.DictWriter(csv_file, fieldnames=self.fieldnames)
            csv_writer.writeheader()
            csv_writer.writerow(radon_findings)



import re
from django import template

register = template.Library()

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


@register.filter(name='to_snake_case')
def to_snake_case(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()



import os

import json
from django.conf import settings
from django.core.management import BaseCommand

from django_test_tools.excel import ExcelAdapter
from django_test_tools.file_utils import add_date, create_dated


class Command(BaseCommand):
    """
        $ python manage.py -i path/to/excel.xlsx -d ./output/
    """

    def add_arguments(self, parser):
        # parser.add_argument('requirement_filename')
        parser.add_argument("-i", "--input",
                            dest="input",
                            help="Input Excel file",
                            )

        parser.add_argument("-o", "--output",
                            dest="output",
                            help="Output filename",
                            )

        # parser.add_argument("--office",
        #                     dest="office",
        #                     help="Organizational unit short name",
        #                     default=None,
        #                     )

        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        dict_data = ExcelAdapter.convert_to_dict(options['input'])
        out_file = options.get('output', create_dated('excel_to_json.json'))
        with open(out_file, 'w') as json_file:
            json.dump(dict_data, json_file, indent=4)

        self.stdout.write('Wrote {}'.format(out_file))



import os

from django.core.management import BaseCommand

from ...file_utils import add_date
from ...flake8.parsers import Flake8Parser, RadonParser


class Command(BaseCommand):
    """

        $ python manage.py
    """

    def add_arguments(self, parser):
        parser.add_argument("-m", "--mode", dest="mode", default="FLAKE8", help="FLAKE8 or RADON")
        parser.add_argument("-i", "--input", dest="input_file", help="Input file")
        parser.add_argument("-o", "--output", dest="output_file", help="Output file")

        # parser.add_argument('output_filename')
        # parser.add_argument("-l", "--list",
        #                     action='store_true',
        #                     dest="list",
        #                     help="List employees",
        #                     )
        # parser.add_argument("-a", "--assign",
        #                     action='store_true',
        #                     dest="assign",
        #                     help="Create unit assignments",
        #                     )
        #
        # parser.add_argument("--office",
        #                     dest="office",
        #                     help="Organizational unit short name",
        #                     default=None,
        #                     )
        # parser.add_argument("--start-date",
        #                     dest="start_date",
        #                     help="Start date for the assignment",
        #                     default=None,
        #                     )
        # parser.add_argument("--fiscal-year",
        #                     dest="fiscal_year",
        #                     help="Fiscal year for assignments",
        #                     default=None,
        #                     )
        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        if not options.get('output_file'):
            output_filename = add_date(options.get('input_file'))
            filename, file_extension = os.path.splitext(output_filename)
            if file_extension != 'csv':
                output_filename = '{}.csv'.format(filename)
        else:
            output_filename = options.get('output_file')

        if options.get('mode') == 'FLAKE8':
            parser = Flake8Parser()
            parser.write_summary(options['input_file'], output_filename)
        elif options.get('mode') == 'RADON':
            parser = RadonParser()
            parser.write_totals(options['input_file'], output_filename)

        self.stdout.write('mode {}'.format(options.get('mode')))
        self.stdout.write('input: {}'.format(options.get('input_file')))
        self.stdout.write('output: {}'.format(output_filename))



import functools
import json
import os
import re
import time
import warnings
from datetime import datetime, date, timedelta

import pytz
from django.conf import settings
from django.utils import timezone
from openpyxl.compat import deprecated

from .file_utils import add_date

__author__ = 'lberrocal'


def versiontuple(v):
    return tuple(map(int, (v.split("."))))

def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o: (d1[o], d2[o]) for o in intersect_keys if d1[o] != d2[o]}
    same = set(o for o in intersect_keys if d1[o] == d2[o])
    return added, removed, modified, same


first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def convert_to_snake_case(camel_case):
    """
    Converts a CamelCase name to snake case.
    ..code-block:: python

        camel_case = 'OperatingSystemLongName'
        snake_case = convert_to_snake_case(camel_case)
        self.assertEqual(snake_case, 'operating_system_long_name')

    :param camel_case: string. Camel case name
    :return: string. Snake case name
    """
    s1 = first_cap_re.sub(r'\1_\2', camel_case)
    return all_cap_re.sub(r'\1_\2', s1).lower()


@deprecated('Should use django_test_tools.file_utils.create_dated() function')
def create_output_filename_with_date(filename):
    """
    Based on the filename will create a full path filename incluidn the date and time in '%Y%m%d_%H%M' format.
    The path to the filename will be set in the TEST_OUTPUT_PATH settings variable.

    :param filename: base filename. my_excel_data.xlsx for example
    :return: string, full path to file with date and time in the TEST_OUTPUT_PATH folder
    """
    if getattr(settings, 'TEST_OUTPUT_PATH', None) is None:
        msg = 'You need a the variable TEST_OUTPUT_PATH in settings. It should point to a folder' \
              'for temporary data to be written and reviewed.'
        raise ValueError(msg)
    if not os.path.exists(settings.TEST_OUTPUT_PATH):
        os.makedirs(settings.TEST_OUTPUT_PATH)
    return add_date(os.path.join(settings.TEST_OUTPUT_PATH, filename))


@deprecated('Should use django_test_tools.file_utils.add_date() function')
def add_date_to_filename(filename, **kwargs):
    """
    Adds to a filename the current date and time in '%Y%m%d_%H%M' format.
    For a filename /my/path/myexcel.xlsx the function would return /my/path/myexcel_20170101_1305.xlsx.
    If the file already exists the function will add seconds to the date to attempt to get a unique name.

    :param filename: string with fullpath to file or just the filename
    :param kwargs: dictionary. date_position: suffix or preffix, extension: string to replace extension
    :return: string with full path string incluiding the date and time
    """
    current_datetime = timezone.localtime(timezone.now()).strftime('%Y%m%d_%H%M%S')
    new_filename_data = dict()
    suffix_template = '{path}{separator}{filename_with_out_extension}_{datetime}.{extension}'
    prefix_template = '{path}{separator}{datetime}_{filename_with_out_extension}.{extension}'
    if '/' in filename and '\\' in filename:
        raise ValueError('Filename %s contains both / and \\ separators' % filename)
    if '\\' in filename:
        path_parts = filename.split('\\')
        file = path_parts[-1]
        path = '\\'.join(path_parts[:-1])
        separator = '\\'
    elif '/' in filename:
        path_parts = filename.split('/')
        file = path_parts[-1]
        path = '/'.join(path_parts[:-1])
        separator = '/'
    else:
        file = filename
        path = ''
        separator = ''

    new_filename_data['path'] = path
    parts = file.split('.')
    if kwargs.get('extension', None) is not None:
        new_filename_data['extension'] = kwargs['extension']
    else:
        new_filename_data['extension'] = parts[-1]

    new_filename_data['separator'] = separator
    new_filename_data['filename_with_out_extension'] = '.'.join(parts[:-1])
    new_filename_data['datetime'] = current_datetime[:-2]
    date_position = kwargs.get('date_position', 'suffix')
    if date_position == 'suffix':
        new_filename = suffix_template.format(**new_filename_data)
        if os.path.exists(new_filename):
            new_filename_data['datetime'] = current_datetime
            new_filename = suffix_template.format(**new_filename_data)
    else:
        new_filename = prefix_template.format(**new_filename_data)
        if os.path.exists(new_filename):
            new_filename_data['datetime'] = current_datetime
            new_filename = prefix_template.format(**new_filename_data)

    return new_filename


def deprecated(func):
    '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.
       from: https://wiki.python.org/moin/PythonDecoratorLibrary#CA-92953dfd597a5cffc650d5a379452bb3b022cdd0_7
    '''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit("Call to deprecated function {}.".format(func.__name__),
                               category=DeprecationWarning,
                               filename=func.__code__.co_filename,
                               lineno=func.__code__.co_firstlineno + 1
                               )
        return func(*args, **kwargs)

    return new_func


@deprecated
def daterange(start_date, end_date):
    """
    DEPRECATED use utils.weekdays() function instead
    :param start_date:
    :param end_date:
    :return:
    """
    return weekdays(start_date, end_date)


def weekdays(start_date, end_date):
    """
    Returns a generator with the dates of the week days between the start and end date

    .. code-block:: python

        start_date = datetime.date(2016, 10, 3)  # Monday
        end_date = datetime.date(2016, 10, 7)  # Friday
        days = list(weekdays(start_date, end_date))
        self.assertEqual(5, len(days))

    :param start_date: date. Start date
    :param end_date: date. End date
    """
    weekend = set([5, 6])
    for n in range(int((end_date - start_date).days) + 1):
        dt = start_date + timedelta(n)
        if dt.weekday() not in weekend:
            yield dt
        else:
            continue


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def force_date_to_datetime(unconverted_date, tzinfo=pytz.UTC):
    converted_datetime = date(year=unconverted_date.year,
                              month=unconverted_date.month,
                              day=unconverted_date.day,
                              hour=0,
                              minute=0,
                              second=0,
                              tzinfo=tzinfo)
    return converted_datetime


class Timer:
    """
    Class to measure time elapsed

    Example:

    .. code:: python

            def test_performance(self):
                with Timer() as stopwatch:
                    web_service = WebServiceUtil()
                    web_service.consume_date(12)
                elapsed_milliseconds = stopwatch.elapsed*1000
                logger.debug('Elapsed: {} ms'.format(elapsed_milliseconds))
                self.assertTrue(elapsed_milliseconds <= 500)

    """

    def __init__(self):
        self.elapsed = 0.0
        self._start = None

    def start(self):
        if self._start is not None:
            raise RuntimeError('Already started')
        self._start = time.perf_counter()

    def stop(self):
        if self._start is None:
            raise RuntimeError('Not started')
        end = time.perf_counter()
        self.elapsed += end - self._start
        self._start = None

    def reset(self):
        self.elapsed = 0.0

    def get_elapsed_time(self):
        hours, remainder = divmod(self.elapsed, 3600)
        mins, secs = divmod(remainder, 60)
        return int(hours), int(mins), secs

    def get_elapsed_time_str(self):
        return '%d h %d m %.2f s' % self.get_elapsed_time()

    @property
    def running(self):
        return self._start is not None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()


def load_json_file(filename):
    with open(filename, 'r') as f:
        data = json.load(f)
    return data


def datetime_to_local_time(date_time):
    """
    Converts a naive date to a time zoned date based in hte setting.TIME_ZONE variable. If the date has already a time zone
    it will localize the date.
    :param date_time: <date> or <datetime> to be localized
    :return: localized non naive datetime
    """
    if isinstance(date_time, date) and not isinstance(date_time, datetime):
        date_time = datetime.combine(date_time, datetime.min.time())

    is_naive = date_time.tzinfo is None or date_time.tzinfo.utcoffset(date_time) is None
    time_zone = pytz.timezone(settings.TIME_ZONE)
    if is_naive:
        return time_zone.localize(date_time)
    else:
        return date_time.astimezone(time_zone)


class SpanishDate(object):
    def __init__(self):
        self.spanish_months = {'Ene': 'Jan', 'Feb': 'Feb', 'Mar': 'Mar', 'Abr': 'Apr', 'May': 'May',
                               'Jun': 'Jun', 'Jul': 'Jul', 'Ago': 'Aug', 'Sep': 'Sep', 'Oct': 'Oct',
                               'Nov': 'Nov', 'Dic': 'Dec'}
        self.english_months = dict()
        for k, v in self.spanish_months.items():
            self.english_months[v] = k

        self.date_format = '%d-%b-%y'
        self.datetime_format = '%d-%b-%y %H:%M'

        month_reg = '|'.join(self.spanish_months.values())
        regex = r'([0123][0-9])-(' + month_reg + ')-([\d]{2})$'
        self.date_reg_exp_en = re.compile(regex)

        regex = regex[:-1] + r' ([012][0-9]:[0-5][0-9])'
        self.datetime_reg_exp_en = re.compile(regex)

        month_reg = '|'.join(self.spanish_months.keys())
        regex = r'([0123][0-9])-(' + month_reg + ')-([\d]{2})$'
        self.date_reg_exp_es = re.compile(regex)

        regex = regex[:-1] + r' ([012][0-9]:[0-5][0-9])'
        self.datetime_reg_exp_es = re.compile(regex)

    def _get_date_parts(self, match, input_lang='es'):
        day = match.group(1)
        if input_lang == 'es':
            month = self.spanish_months[match.group(2)]
        else:
            month = self.english_months[match.group(2)]
        year = match.group(3)
        return day, month, year

    def parse(self, str_date):
        match = self.date_reg_exp_es.match(str_date)
        date_to_parse = None
        if match:
            day, month, year = self._get_date_parts(match)
            english_date = '{}-{}-{}'.format(day, month, year)
            date_to_parse = datetime_to_local_time(datetime.strptime(english_date, self.date_format)).date()
        else:
            match = self.datetime_reg_exp_es.match(str_date)
            if match:
                day, month, year = self._get_date_parts(match)
                time = match.group(4)
                english_date = '{}-{}-{} {}'.format(day, month, year, time)
                date_to_parse = datetime_to_local_time(datetime.strptime(english_date, self.datetime_format))
        return date_to_parse

    def to_string(self, m_date):

        if isinstance(m_date, datetime):
            str_date = m_date.strftime(self.datetime_format)
            match = self.datetime_reg_exp_en.match(str_date)
            if match:
                day, month, year = self._get_date_parts(match, 'en')
                time = match.group(4)
                return '{}-{}-{} {}'.format(day, month, year, time)
        elif isinstance(m_date, date):
            str_date = m_date.strftime(self.date_format)
            match = self.date_reg_exp_en.match(str_date)
            # date = None
            if match:
                day, month, year = self._get_date_parts(match, 'en')
                return '{}-{}-{}'.format(day, month, year)
        return None


spanish_date_util = SpanishDate()


def parse_spanish_date(str_date):
    return spanish_date_util.parse(str_date)



from django.core.checks import register, Error, Tags


@register(Tags.compatibility, deploy=True)
def example_check(app_configs, **kwargs):
    errors = []
    check_failed = True

    if check_failed:
        errors.append(
            Error(
                'an error',
                hint='A hint.',
                obj='kKKK',
                id='myapp.E001',
            )
        )
    return errors



import re


class CommonRegExp(object):
    """
    This class is a utility for commonly used regular expressions.

    .. code-block:: python

        str_data = '10:45'
        common_regexp = CommonRegExp()
        regular_expression, key = common_regexp.match_regexp(str_data)
        self.assertEqual(regular_expression, '([0-1][0-9]|2[0-4]):([0-5][0-9])')
        self.assertEqual(key, 'time_military')

    if the strict keyword is used the class will add ^ at the begining and a $ at the end if the already are not present.

    """

    def __init__(self,**kwargs):
        self.strict = kwargs.get('strict', False)
        self.regular_expressions = dict()
        self.regular_expressions['time_military'] = {
            'pattern': r'([0-1][0-9]|2[0-4]):([0-5][0-9])'
        }
        self.regular_expressions['integer'] = {
            'pattern': r'^(\d+)$'
        }
        self.regular_expressions['decimal'] = {
            'pattern': r'(\d+\.\d+)'
        }
        self.regular_expressions['url'] = {
            'pattern': r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
        }
        if kwargs.get('reg_exps'):
            self.regular_expressions.update(kwargs.get('reg_exps'))

        self.regular_expressions_compiled = dict()
        for key, reg_exp in self.regular_expressions.items():
            #self.regular_expressions_compiled[key] = re.compile(reg_exp['pattern'])
            self.regular_expressions_compiled[key] = self._compile_regular_expression(self.regular_expressions[key])

    def match_regexp(self, value):
        for key, regexp in self.regular_expressions_compiled.items():
            match = regexp.match(value)
            if match:
                return regexp.pattern, key

        return None, None

    def add_regular_expression(self, name, pattern, **kwargs):
        regexp_dict = dict()
        regexp_dict['pattern'] = pattern
        if kwargs.get('strict') is not None:
            regexp_dict['strict'] = kwargs.get('strict')
        self.regular_expressions_compiled[name] = self._compile_regular_expression(regexp_dict)

    def _compile_regular_expression(self, regexp_dict):
        if (self.strict or regexp_dict.get('strict')) and not regexp_dict['pattern'].startswith('^'):
            regexp_dict['pattern'] = '^{}'.format(regexp_dict['pattern'])

        if (self.strict or regexp_dict.get('strict')) and not regexp_dict['pattern'].endswith('$'):
            regexp_dict['pattern'] = '{}$'.format(regexp_dict['pattern'])
        return re.compile(regexp_dict['pattern'])




from django.apps.registry import apps


class DjangoAppManager(object):
    def __init__(self):
        self.installed_apps = dict(self.get_installed_apps())

    def get_app(self, app_name):
        return self.installed_apps.get(app_name)

    def get_installed_apps(self):
        for app_config in apps.get_app_configs():
            yield app_config.name, app_config

    def get_model(self, app_name, model_name):
        app = self.get_app(app_name)
        for model in app.get_models():
            if model_name == model.__name__:
                return model
        return None

    def get_project_apps(self, project_name):
        project_apps = dict()
        for app_name, app_config in self.installed_apps.items():
            app_project = app_name.split('.')[0]
            if app_project == project_name:
                project_apps[app_name] = app_config
        return project_apps

    def get_app_data(self, app_name):
        """
        Read application data converts into a dictionary

        :param app_name: Application name
        :return: Dictionary with application data
        """
        app = self.get_app(app_name)
        app_dict = dict()
        app_dict['app_name'] = app.name
        app_dict['models'] = dict()
        for key, model in app.models.items():
            app_dict['models'][key] = dict()
            app_dict['models'][key]['model_name'] = model.__name__
            app_dict['models'][key]['original_attrs'] = model._meta.original_attrs
            app_dict['models'][key]['fields'] = list()
            for field in model._meta.fields:
                self._convert_field_to_dict(field, app_dict['models'][key]['fields'])

        return app_dict

    def _convert_field_to_dict(self, field, field_list):
        field_dict = dict()
        field_dict['field_name'] = field.name
        field_dict['type'] = type(field).__name__
        field_dict['unique'] = field.unique
        field_dict['editable'] = field.editable
        if hasattr(field, 'choices') and field.choices is not None:
            field_dict['choices_type'] = type(field.choices).__name__
            if field_dict['choices_type'] == 'tuple':
                field_dict['choices'] = field.choices
        if hasattr(field, 'max_length') and field.max_length is not None:
            field_dict['max_length'] = field.max_length
        if hasattr(field, 'max_digits') and field.max_digits is not None:
            field_dict['max_digits'] = field.max_digits
        if hasattr(field, 'decimal_places') and field.decimal_places is not None:
            field_dict['decimal_places'] = field.decimal_places
        if hasattr(field, 'remote_field') and field.remote_field is not None:
            field_dict['remote_field'] = field.remote_field.model.__name__
        field_list.append(field_dict)
        return field_dict



import os

from django.conf import settings
from django.core.management import BaseCommand

from django_test_tools.generators.crud_generator import GenericTemplateWriter
from django_test_tools.generators.model_generator import ModelSerializerGenerator
from ...generators.serializer_gen import AppSerializerGenerator
from ...app_manager import DjangoAppManager


class Command(BaseCommand):
    """

        $ python manage.py generate_serializers project.app -s ModelSerializer
    """

    def add_arguments(self, parser):
        parser.add_argument('app_name')
        parser.add_argument("-s", "--serializer-class",
                            dest="serializer_class",
                            help="Serializer class",
                            default='ModelSerializer'
                            )
        parser.add_argument('-f', "--filename",
                            dest="filename",
                            help="Output filename. Will write it to the settings.TEST_OUTPUT_PATH folder",
                            default=None,
                            )
        # parser.add_argument("-a", "--assign",
        #                     action='store_true',
        #                     dest="assign",
        #                     help="Create unit assignments",
        #                     )
        #
        # parser.add_argument("--office",
        #                     dest="office",
        #                     help="Organizational unit short name",
        #                     default=None,
        #                     )
        # parser.add_argument("--start-date",
        #                     dest="start_date",
        #                     help="Start date for the assignment",
        #                     default=None,
        #                     )
        # parser.add_argument("--fiscal-year",
        #                     dest="fiscal_year",
        #                     help="Fiscal year for assignments",
        #                     default=None,
        #                     )
        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        app_name = options.get('app_name', None)
        if options.get('filename'):
            filename = os.path.join(settings.TEST_OUTPUT_PATH, options.get('filename'))
            generator = ModelSerializerGenerator()
            factory_data = generator.create_template_data(app_name)
            template_name = 'serializers.py.j22'
            writer = GenericTemplateWriter(template_name)
            writer.write(factory_data, filename)
        else:
            serializer_class = options.get('serializer_class')
            app_manager = DjangoAppManager()
            app = app_manager.get_app(app_name)
            if app:
                app_model_tests = AppSerializerGenerator(app, serializer_class)
                self.stdout.write(str(app_model_tests))



import csv
import json
import os
from io import StringIO

from django.conf import settings
from django.urls import reverse
from rest_framework.test import APIClient

from .excel import ExcelAdapter
from .exceptions import DjangoTestToolsException


class TestCommandMixin(object):
    """
    This mixin helps capture the output of a command written with the stdout.write() method and
     the stderr.write

    .. code-block:: python

        class TestYourCommand(TestCommandMixin, TestCase):

            def test_your_command_action(self):
                call_command('your_command', 'your_argument', stdout=self.content, stderr=self.error_content)
                results = self.get_results()
                self.assertEqual(23, len(results))
    """

    # noinspection PyPep8Naming
    def setUp(self):
        self.content = StringIO()
        self.error_content = StringIO()

    def get_results(self, content=None):
        if content is None:
            content = self.content
        content.seek(0)
        lines = content.readlines()
        results = list()
        for line in lines:
            results.append(line.strip('\n'))
        return results

    def get_errors(self):
        return self.get_results(self.error_content)


class TestOutputMixin(object):
    clean_output = True

    def clean_output_folder(self, dated_filename):
        if self.clean_output:
            os.remove(dated_filename)
            # noinspection PyUnresolvedReferences
            self.assertFalse(os.path.exists(dated_filename))

    def get_excel_content(self, filename, sheet_name=None):
        """
        Reads the content of an excel file and returns the content a as list of row lists.
        :param filename: string full path to the filename
        :param sheet_name: string. Name of the sheet to read if None will read the active sheet
        :return: a list containing a list of values for every row.
        """
        adapter = ExcelAdapter()
        return adapter.convert_to_list(filename, sheet_name)

    def get_csv_content(self, filename, delimiter=',', encoding='utf-8'):
        content = list()
        with open(filename, 'r', encoding=encoding) as csv_file:
            reader = csv.reader(csv_file, delimiter=delimiter)
            for row in reader:
                content.append(row)
        return content

    def get_txt_content(self, filename, encoding='utf-8'):
        content = list()
        with open(filename, 'r', encoding=encoding) as file:
            lines = file.readlines()
        for line in lines:
            content.append(line.strip('\n'))
        return content


class JWTTestMixin(object):

    def get_access_token(self, user):
        token_url = reverse('token_obtain_pair')
        pay_load = {'username': user.username, 'password': 'password'}
        token_response = self.post(token_url, data=pay_load)
        access_token = token_response.data['access']
        return access_token

    def get_with_token(self, url, access_token):
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(access_token))
        response = client.get(url, data={'format': 'json'})
        return response

    def delete_with_token(self, url, access_token):
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(access_token))
        response = client.delete(url, data={'format': 'json'})
        return response

    def put_with_token(self, url, access_token, data):
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(access_token))
        response = client.put(url, data=data)
        return response


class TestFixtureMixin(object):
    """
    This a mixin to add to test cases to easily access fixtures. It assumes the the you have a package for your tests
    named **tests** and your **fixtures** are in a folder named fixtures within your tests package and that you have
    settings variable named APPS_DIR pointing tou your applications folder (this is created by Cookiecutter by default).
    Your tests package should look like this.

    |   ├── clinics
    │   ├── __init__.py
    │   ├── admin.py
    │   ├── apps.py
    │   ├── exceptions.py
    │   ├── forms.py
    │   ├── migrations
    │   ├── models.py
    │   ├── tests
    │   │   ├── __init__.py
    │   │   ├── factories.py
    │   │   ├── fixtures
    │   │   │   ├── data.json
    │   │   │   ├── model_data.txt
    │   │   ├── test_forms.py
    │   │   ├── test_models.py
    │   ├── urls.py
    │   └── views.py

    For the above exmple

    .. code-block:: python

        class TestClinicAdapter(TestFixtureMixin, TestCase):

            def setUp(self) -> None:
                self.app_name = 'clinics'

            def test_parse(self):
                clinics_dictionary = self.get_fixture_json('data.json')
                ...

            def test_read(self):
                filename = self.get_fixture_fullpath('model.txt')
                ...

    """
    app_name = None

    def __init__(self, app_name=None, **kwargs):
        self.app_name = app_name
        ## this is to test and allow app_name to be None.
        self.strict = kwargs.get('strict', True)

    def get_fixture_fullpath(self, fixture_filename):
        """
        Get full patch for the fixture file

        :param fixture_filename: <str> name of the fixture file
        :return: <str> full path to fixture file
        """
        self._sanity_check()
        if self.app_name is None:
            fixture_file = settings.APPS_DIR.path('tests', 'fixtures', fixture_filename).root
        else:
            fixture_file = settings.APPS_DIR.path(self.app_name, 'tests', 'fixtures', fixture_filename).root
        return fixture_file

    def get_fixture_json(self, fixture_filename):
        """
        Reads a file and returns the json content.

        :param fixture_filename: <str> filename
        :return: <dict> With the file content
        """
        fixture_file = self.get_fixture_fullpath(fixture_filename)

        with open(fixture_file, 'r') as f:
            json_data = json.load(f)
        return json_data

    def _sanity_check(self):
        if self.app_name is None and self.strict:
            raise DjangoTestToolsException('app_name not defined')
        if not hasattr(settings, 'APPS_DIR'):
            raise DjangoTestToolsException('APPS_DIR not defined')



import collections
from datetime import datetime, date
from decimal import Decimal

from openpyxl.compat import deprecated

from .re.regexp import CommonRegExp
from .utils import create_output_filename_with_date


def write_assertions(dictionary_list, variable_name, **kwargs):
    """
    Writes assertions using Django practice of putting actual value first and then expected value to a file.
    If no filename is supplied it will generate a file in the settings.TEST_OUTPUT_PATH folder with the
    **variable_name** and the current date.
    By default key named created and modified will be excluded.

    :param dictionary_list: <list> or <dict> dictionary or list of values
    :param variable_name: <str> name of the variable
    :param kwargs:  filename <str>String. Full path to the output file.
    :param kwargs:  excluded_keys <list>list of strings. List with keys to exclude
    :param kwargs:  type_only <boolean> Check only for types instead of values. Default false
    :return: filename string.
    """
    writer = AssertionWriter(**kwargs)
    return writer.write_assert_list(dictionary_list, variable_name, filename=kwargs.get('filename'))


@deprecated('Use assert_utils.write_assertions instead')
def write_assert_list(filename, dictionary_list, variable_name):
    """
    Function to generate assertions for a dictionary or list content.
    :param filename:
    :param dictionary_list:
    :param variable_name:
    :return:
    """
    writer = AssertionWriter()
    return writer.write_assert_list(dictionary_list, variable_name, filename=filename)


class AssertionWriter(object):
    """
    This class generates assertions using Django practice of putting actual value first and then expected value.
    """

    def __init__(self, **kwargs):
        self.excluded_variable_names = ['created', 'modified']
        if kwargs.get('excluded_keys') is not None:
            for key in kwargs.get('excluded_keys'):
                self.excluded_variable_names.append(key)

        self.use_regexp_assertion = kwargs.get('use_regexp_assertion', False)
        if self.use_regexp_assertion:
            self.common_regexp = CommonRegExp(strict=True)
        self.check_for_type_only = kwargs.get('type_only', False)

    def add_regular_expression(self, name, pattern, **kwargs):
        self.common_regexp.add_regular_expression(name, pattern, **kwargs)

    def write_assert_list(self, dictionary_list, variable_name, **kwargs):
        """
        Function to generate assertions for a dictionary or list content.
        :param kwargs:
        :param dictionary_list:
        :param variable_name:
        :return:
        """
        if kwargs.get('filename') is None:
            filename = create_output_filename_with_date('{}.py'.format(variable_name))
        else:
            filename = kwargs.get('filename')
        if self.check_for_type_only:
            if isinstance(dictionary_list, dict):
                assert_list = self._generate_assert_type_dictionaries(dictionary_list, variable_name)
            elif isinstance(dictionary_list, list):
                assert_list = self._generate_assert_type_list(dictionary_list, variable_name)
        else:
            if isinstance(dictionary_list, dict):
                assert_list = self._generate_assert_equals_dictionaries(dictionary_list, variable_name)
            elif isinstance(dictionary_list, list):
                assert_list = self._generate_assert_equals_list(dictionary_list, variable_name)

        with open(filename, 'w', encoding='utf-8', newline='\n') as python_file:
            python_file.write('\n'.join(assert_list))

        return filename

    def _generate_assert_equals_list(self, data_list, variable_name, indentation_level=0):
        assert_list = list()
        if variable_name not in self.excluded_variable_names:
            index = 0
            # assert_list.append('# ********** variable {} ***********'.format(variable_name))
            assert_list.append('self.assertEqual({}, {})'.format('len({})'.format(variable_name), len(data_list)))
            for data in data_list:
                list_variable = '{}[{}]'.format(variable_name, index)
                self._build_equals_assertion(list_variable, data, assert_list)
                index += 1
        return assert_list

    def _generate_assert_equals_dictionaries(self, dictionary, variable_name, **kwargs):
        assert_list = list()
        if variable_name not in self.excluded_variable_names:
            ordered_dictionary = collections.OrderedDict(sorted(dictionary.items()))
            for key, value in ordered_dictionary.items():
                if key not in self.excluded_variable_names:
                    dict_variable = '{}[\'{}\']'.format(variable_name, key)
                    self._build_equals_assertion(dict_variable, value, assert_list)
        return assert_list

    def _build_equals_assertion(self, variable_name, data, assert_list):
        if variable_name not in self.excluded_variable_names:
            if isinstance(data, str):
                data = data.translate(str.maketrans({"'": '\\\''}))
                if self.use_regexp_assertion:
                    pattern = self.common_regexp.match_regexp(data)[0]
                    if pattern is None:
                        assert_list.append('self.assertEqual({}, \'{}\')'.format(variable_name, data))
                    else:
                        assert_list.append('self.assertRegex({}, r\'{}\')'.format(variable_name, pattern))
                else:
                    assert_list.append('self.assertEqual({}, \'{}\')'.format(variable_name, data))
            elif isinstance(data, datetime):
                date_time_format = '%Y-%m-%d %H:%M:%S%z'
                str_datetime = data.strftime(date_time_format)
                assert_list.append(
                    'self.assertEqual({}.strftime(\'{}\'), \'{}\')'.format(
                        variable_name,
                        date_time_format,
                        str_datetime,
                    )
                )
            elif isinstance(data, date):
                date_format = '%Y-%m-%d'
                str_date = data.strftime(date_format)
                assert_list.append(
                    'self.assertEqual({}.strftime(\'{}\'), \'{}\')'.format(variable_name, date_format, str_date))
            elif isinstance(data, Decimal):
                assert_list.append('self.assertEqual({}, Decimal({}))'.format(variable_name, data))
            elif isinstance(data, list):
                assert_list += self._generate_assert_equals_list(data, variable_name)
            elif isinstance(data, dict):
                assert_list += self._generate_assert_equals_dictionaries(data, variable_name)
            else:
                assert_list.append('self.assertEqual({}, {})'.format(variable_name, data))

    def _build_type_assertion(self, variable_name, data, assert_list):
        if variable_name not in self.excluded_variable_names:
            if isinstance(data, str):
               assert_list.append('self.assertIsNotNone({}) # Example: {}'.format(variable_name, data))
            elif isinstance(data, datetime):
                date_time_format = '%Y-%m-%d %H:%M:%S%z'
                datetime_regexp = r'([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))\s\d{2}:\d{2}:\d{2}\+\d{4}'
                str_datetime = data.strftime(date_time_format)
                assert_list.append(
                    'self.assertRegex({}.strftime(\'{}\'), r\'{}\') # Example: {}'.format(
                        variable_name,
                        date_time_format,
                        datetime_regexp,
                        str_datetime,
                    )
                )
            elif isinstance(data, date):
                date_format = '%Y-%m-%d'
                date_regexp = r'([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))'
                str_date = data.strftime(date_format)
                assert_list.append(
                    'self.assertRegex({}.strftime(\'{}\'), r\'{}\') # Example: {}'.format(
                        variable_name,
                        date_format,
                        date_regexp,
                        str_date))
            elif isinstance(data, Decimal):
                assert_list.append('self.assertIsNotNone({}) # Example: Decimal({})'.format(variable_name, data))
            elif isinstance(data, list):
                assert_list += self._generate_assert_type_list(data, variable_name)
            elif isinstance(data, dict):
                assert_list += self._generate_assert_type_dictionaries(data, variable_name)
            else:
                if data is None:
                    assert_list.append('self.assertIsNone({}) # Example: {}'.format(variable_name, data))
                else:
                    assert_list.append('self.assertIsNotNone({}) # Example: {}'.format(variable_name, data))

    def _generate_assert_type_dictionaries(self, dictionary, variable_name, **kwargs):
        assert_list = list()
        assert_list.append('self.assertEqual(len({}.keys()), {})'.format(variable_name, len(dictionary.keys())))
        if variable_name not in self.excluded_variable_names:
            ordered_dictionary = collections.OrderedDict(sorted(dictionary.items()))
            for key, value in ordered_dictionary.items():
                if key not in self.excluded_variable_names:
                    dict_variable = '{}[\'{}\']'.format(variable_name, key)
                    self._build_type_assertion(dict_variable, value, assert_list)
        return assert_list

    def _generate_assert_type_list(self, data_list, variable_name, indentation_level=0):
        assert_list = list()
        if variable_name not in self.excluded_variable_names:
            index = 0
            # assert_list.append('# ********** variable {} ***********'.format(variable_name))
            assert_list.append('self.assertEqual({}, {})'.format('len({})'.format(variable_name), len(data_list)))
            for data in data_list:
                list_variable = '{}[{}]'.format(variable_name, index)
                self._build_type_assertion(list_variable, data, assert_list)
                index += 1
        return assert_list



from django.template.loader import render_to_string
from jinja2 import Environment, PackageLoader, select_autoescape

from ..templatetags.dtt_filters import to_snake_case


class UrlGenerator(object):

    def __init__(self, model_name):
        self.model_name = model_name
        self.template = 'django_test_tools/urls.py.j2'

    def print_urls(self, filename):
        self._print(filename, 'urls')

    def print_paths(self, filename):
        self._print(filename, 'paths')

    def _print(self, filename, what_to_print):
        data = dict()
        data['model_name'] = self.model_name
        data['print_{}'.format(what_to_print)] = True
        rendered = render_to_string(self.template, data)
        with open(filename, 'w', encoding='utf-8') as url_file:
            url_file.write(rendered)


class SerializerTestGenerator(object):

    def __init__(self):
        self.env = Environment(
            loader=PackageLoader('django_test_tools', 'templates'),
            autoescape=select_autoescape(['html', 'j2'])
        )
        self.env.filters['to_snake_case'] = to_snake_case

        self.template_name = 'django_test_tools/test_serializers.py.j2'
        self.template = self.env.get_template(self.template_name)

    def print(self, serializer_info, filename):
        rendered = self.template.render(serializer_info)
        with open(filename, 'w', encoding='utf-8') as file:
            file.write(rendered)


class GenericTemplateWriter(object):

    def __init__(self, template_name, **kwargs):
        trim_blocks = kwargs.get('trim_blocks', False)
        lstrip_blocks = kwargs.get('lstrip_blocks', False)
        self.env = Environment(
            loader=PackageLoader('django_test_tools', 'templates'),
            autoescape=select_autoescape(['html', ],),
            trim_blocks=trim_blocks,
            lstrip_blocks=lstrip_blocks
        )
        self.env.filters['to_snake_case'] = to_snake_case

        self.template_name = 'django_test_tools/{}'.format(template_name)
        self.template = self.env.get_template(self.template_name)

    def write(self, template_data, filename):
        rendered = self.template.render(template_data)
        with open(filename, 'w', encoding='utf-8') as file:
            file.write(rendered)



import contextlib
import subprocess

import pip
import pkg_resources

from django_test_tools.exceptions import DjangoTestToolsException
from django_test_tools.utils import versiontuple

if versiontuple(pip.__version__) >= (10, 0, 0):
    from pip._internal.req import parse_requirements
    from pip._internal.req.req_install import InstallRequirement
else:
    from pip.req import parse_requirements, InstallRequirement

import re

import requests
import logging

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def capture():
    import sys
    from io import StringIO
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


def parse_pip_list(line):
    regexp = re.compile(r'([\w\-]*)\s\(([\w\-_\.]*)\)\s\-\sLatest:\s([\w\-_\.]*)\s\[([a-z]*)\]')
    regexp_new = re.compile(r'([\w\-]*)\s+([\w\-_\.]*)\s+([\w\-_\.]*)\s+([\w\-_\.]*)')
    regexp_list = re.compile(r'([\w\-]*)\s+([\w\-_\.]*)')
    match = regexp.match(line)
    if match is None:
        match = regexp_new.match(line)
    if match:
        library = dict()
        library['name'] = match.group(1).lower()
        library['current_version'] = match.group(2)
        library['new_version'] = match.group(3)
        return library
    match_list = regexp_list.match(line)
    if match_list:
        library = dict()
        library['name'] = match_list.group(1).lower()
        library['current_version'] = match_list.group(2)
        return library

    return None


def get_latest_version(package_name):
    url = 'https://pypi.python.org/pypi/{}/json'.format(package_name)
    r = requests.get(url)
    versions = sorted(r.json()["releases"], key=pkg_resources.parse_version)
    return versions[-1]


def parse_comes_from(comes_from):
    regexp = re.compile(r'(\-r)\s([/\w\.\-]*)\s\(line\s(\d*)\)')
    match = regexp.match(comes_from)
    if match:
        return match.group(1), match.group(2), int(match.group(3))
    else:
        raise ValueError('Invalid comes from "{}"'.format(comes_from))


def parse_specifier(specifier):
    regexp = re.compile(r'((?:[=>])=)([\w\-_\.]*)')
    match = regexp.match(specifier)
    if match:
        return match.group(1), match.group(2)
    else:
        raise ValueError('Invalid speficier "{}"'.format(specifier))


def read_requirement_file(req_file):
    # print('pip version {}'.format(pip.__version__))
    requirements = dict()
    for item in parse_requirements(req_file, session="somesession"):
        if isinstance(item, InstallRequirement):
            requirement = dict()
            requirement['name'] = item.name
            if len(str(item.req.specifier)) > 0:
                operator, version = parse_specifier(str(item.req.specifier))
                requirement['operator'] = operator
                requirement['version'] = version
            requirement['comes_from'] = dict()
            requirement['comes_from']['value'] = item.comes_from
            file_indicator, filename, line_no = parse_comes_from(item.comes_from)
            requirement['comes_from']['file_indicator'] = file_indicator
            requirement['comes_from']['filename'] = filename
            requirement['comes_from']['line_no'] = line_no

            requirements[item.name] = requirement
    return requirements


def list_outdated_libraries():
    return list_libraries(True)


def list_libraries(outdated=False):
    command_list = ['pip', 'list']
    if outdated:
        command_list.append('--outdated')
    pip_list = subprocess.check_output(command_list)
    # pip.main(['list', '--outdated'])
    library_lines = pip_list.decode("utf-8").split('\n')
    outdated_libraries = list()
    for line in library_lines:
        library = parse_pip_list(line)
        if library is not None:
            outdated_libraries.append(library)
    return outdated_libraries


def update_outdated_libraries(requirement_file, **kwargs):
    """
    Updates the requirements to their latest version.

    :param requirement_file: String with the path to the requirement fiel
    :param kwargs:
    :return: list of dictionaries containting the changes
    """
    requirements = read_requirement_file(requirement_file)
    outdated_libraries = list_outdated_libraries()
    changes = list()
    for outdated_library in outdated_libraries:
        library_name = outdated_library['name']
        if requirements.get(library_name):
            change = dict()
            change['library_name'] = library_name
            change['filename'] = requirements[library_name]['comes_from']['filename']
            line_no = requirements[library_name]['comes_from']['line_no'] - 1
            operator = requirements[library_name]['operator']
            with open(change['filename'], 'r') as file:
                data = file.readlines()
            try:
                new_value = '{}{}{}\n'.format(library_name, operator, outdated_library['new_version'])
            except KeyError as e:
                msg = 'Could not find new version for library {library_name} in {filename}'.format(**change)
                raise DjangoTestToolsException(msg)
            change['previous'] = data[line_no].strip('\n')
            change['new'] = new_value.strip('\n')
            data[line_no] = new_value
            with open(change['filename'], 'w') as file:
                file.writelines(data)
            change['line_no'] = line_no
            changes.append(change)
    return changes



from django.core.management import BaseCommand

from ...app_manager import DjangoAppManager
from ...generators.model_test_gen import AppModelsTestCaseGenerator


class Command(BaseCommand):
    """

        $ python manage.py
    """

    def add_arguments(self, parser):
        parser.add_argument('app_name')
        # parser.add_argument("-l", "--list",
        #                     action='store_true',
        #                     dest="list",
        #                     help="List employees",
        #                     )
        # parser.add_argument("-a", "--assign",
        #                     action='store_true',
        #                     dest="assign",
        #                     help="Create unit assignments",
        #                     )
        #
        # parser.add_argument("--office",
        #                     dest="office",
        #                     help="Organizational unit short name",
        #                     default=None,
        #                     )
        # parser.add_argument("--start-date",
        #                     dest="start_date",
        #                     help="Start date for the assignment",
        #                     default=None,
        #                     )
        # parser.add_argument("--fiscal-year",
        #                     dest="fiscal_year",
        #                     help="Fiscal year for assignments",
        #                     default=None,
        #                     )
        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        app_name = options.get('app_name', None)
        app_manager = DjangoAppManager()
        app = app_manager.get_app(app_name)
        if app:
            app_model_tests = AppModelsTestCaseGenerator(app)
            self.stdout.write(str(app_model_tests))



from django.utils import timezone

from django_test_tools import __version__ as current_version
from ..app_manager import DjangoAppManager


class FactoryBoyGenerator(object):
    """
    Currently supporting:

        - BooleanField
        - CharField
        - CountryField: from https://github.com/SmileyChris/django-countries
        - DateField:
        - DateTimeField
        - DecimalField
        - ForeignKey
        - GenericIPAddressField
        - IntegerField
        - MoneyField: from https://github.com/django-money/django-money
        - TextField

    The following fields are ignored by default 'AutoField', 'AutoCreatedField', 'AutoLastModifiedField'
    """

    def __init__(self, **kwargs):
        self.app_manager = DjangoAppManager()
        self.read_only_fields = ['AutoField', 'AutoCreatedField', 'AutoLastModifiedField']
        self.fields_to_ignore = []
        self.default_countries = kwargs.get('default_countries', ['PA', 'US', 'GB'])
        # names that if are varchar will be filled with digits
        self.numerical_identifiers = kwargs.get('numerical_identifiers', ['id', 'num', ])
        # max_length to consider an alpha numeric code is below or equal a string of letters will be used
        self.code_threshold = kwargs.get('code_threshold', 5)

    def create_template_data(self, app_name):
        app_data = self.app_manager.get_app_data(app_name)
        template_data = dict()
        template_data['models'] = dict()
        template_data['app_name'] = app_data['app_name']
        for model_key in app_data['models'].keys():
            template_data['models'][model_key] = dict()
            template_data['models'][model_key]['model_name'] = app_data['models'][model_key]['model_name']
            template_data['models'][model_key]['package_name'] = '{}.{}'.format(app_data['app_name'],
                                                                                app_data['models'][model_key][
                                                                                    'model_name'])

            template_data['models'][model_key]['fields'] = list()
            if app_data['models'][model_key].get('original_attrs'):
                if app_data['models'][model_key].get('original_attrs').get('unique_together'):
                    template_data['models'][model_key]['unique_together'] = app_data['models'][model_key].get(
                        'original_attrs').get('unique_together')

            for field in app_data['models'][model_key]['fields']:
                field_dict = dict()
                field_dict['field_name'] = field['field_name']
                field_dict['type'] = field['type']
                field_dict['is_supported'] = True
                if field['type'] in self.read_only_fields or field['type'] in self.fields_to_ignore:
                    pass
                else:
                    method_name = 'get_{}_factory'.format(field['type'].lower())
                    if hasattr(self, method_name):
                        field_dict['factory'] = getattr(self, method_name)(field, template_data['models'][model_key])
                        template_data['models'][model_key]['fields'].append(field_dict)
                    else:
                        field_dict['is_supported'] = False
                        template_data['models'][model_key]['fields'].append(field_dict)
        return template_data

    def get_booleanfield_factory(self, *args, **kwargs):
        return 'Iterator([True, False])'

    def get_countryfield_factory(self, *args, **kwargs):
        country_list_str = ''
        for country in self.default_countries:
            country_list_str += '\'{}\', '.format(country)
        template = 'Iterator([{}])'
        return template.format(country_list_str.strip())

    def get_charfield_factory(self, *args, **kwargs):
        if args[0].get('choices'):
            template = 'Iterator({choices}, getter=lambda x: x[0])'
            return template.format(**args[0])

        if self._is_number(args[0]['field_name']):
            template = 'LazyAttribute(lambda x: FuzzyText(length={max_length}, chars=string.digits).fuzz())'
            return template.format(**args[0])

        if args[0]['max_length'] > self.code_threshold:
            template = 'LazyAttribute(lambda x: faker.text(max_nb_chars={max_length}))'
            return template.format(**args[0])
        else:
            template = 'LazyAttribute(lambda x: FuzzyText(length={max_length}, chars=string.ascii_letters).fuzz())'
            return template.format(**args[0])

    def get_datefield_factory(self, *args, **kwargs):
        config = {'start_date': kwargs.get('start_date', '-1y'),
                  'end_date': kwargs.get('end_date', 'today')}
        template = 'LazyAttribute(lambda x: faker.date_between(start_date="{start_date}", ' \
                   'end_date="{end_date}", tzinfo=timezone(settings.TIME_ZONE)))'
        return template.format(**config)

    def get_datetimefield_factory(self, *args, **kwargs):
        config = {'start_date': kwargs.get('start_date', '-1y'),
                  'end_date': kwargs.get('end_date', 'now')}
        template = 'LazyAttribute(lambda x: faker.date_time_between(start_date="{start_date}", ' \
                   'end_date="{end_date}", tzinfo=timezone(settings.TIME_ZONE)))'
        return template.format(**config)

    def get_decimalfield_factory(self, *args, **kwargs):
        config = {
            'left_digits': args[0]['max_digits'] - args[0]['decimal_places'],
            'right_digits': args[0]['decimal_places'],
            'positive': kwargs.get('positive', True)
        }
        template = 'LazyAttribute(lambda x: faker.pydecimal(left_digits={left_digits}, ' \
                   'right_digits={right_digits}, positive={positive}))'
        return template.format(**config)

    def get_filefield_factory(self, *args, **kwargs):
        config = {'filename': kwargs.get('filename', args[0]['field_name']),
                  'extension': kwargs.get('extension', 'txt'), }
        template = 'FileField(filename=\'{filename}.{extension}\')'
        return template.format(**config)

    def get_foreignkey_factory(self, *args, **kwargs):
        config = {'model_name': args[0]['remote_field']}
        template = 'SubFactory({model_name}Factory)'
        return template.format(**config)

    def get_genericipaddressfield_factory(self, *args, **kwargs):
        config = {'network': False}
        template = 'LazyAttribute(lambda o: faker.ipv4(network={network}))'
        return template.format(**config)

    def get_integerfield_factory(self, *args, **kwargs):
        config = {'min_value': kwargs.get('min_value', 3),
                  'max_value': kwargs.get('max_value', 3), }
        template = 'LazyAttribute(lambda o: randint({min_value}, {max_value}))'
        return template.format(**config)

    def get_moneyfield_factory(self, *args, **kwargs):
        factory = self.get_decimalfield_factory(*args, **kwargs)
        return factory

    def get_textfield_factory(self, *args, **kwargs):
        config = {'sentences': kwargs.get('sentences', 3)}
        template = 'LazyAttribute(lambda x: faker.paragraph(nb_sentences={sentences}, variable_nb_sentences=True))'
        return template.format(**config)

    def _is_number(self, field_name):
        for nv in self.numerical_identifiers:
            if nv in field_name.lower():
                return True
        return False


class ModelSerializerGenerator(object):

    def __init__(self, **kwargs):
        self.app_manager = DjangoAppManager()
        self.field_types_to_ignore = kwargs.get('field_types_to_ignore', [])

    def create_template_data(self, app_name):
        app_data = self.app_manager.get_app_data(app_name)
        template_data = dict()
        template_data['template_data'] = dict()
        template_data['template_data']['version'] = current_version
        template_data['template_data']['generation_date'] = timezone.now().strftime('%Y-%m-%d %H:%M:%S')
        template_data['template_data']['tz'] = str(timezone.now().tzinfo)

        template_data['models'] = dict()
        template_data['app_name'] = app_data['app_name']
        for model_key in app_data['models'].keys():
            template_data['models'][model_key] = dict()
            template_data['models'][model_key]['model_name'] = app_data['models'][model_key]['model_name']
            template_data['models'][model_key]['package'] = app_data['app_name']
            template_data['models'][model_key]['qualified_name'] = '{}.{}'.format(app_data['app_name'],
                                                                                  app_data['models'][model_key][
                                                                                      'model_name'])
            template_data['models'][model_key]['has_ignored_fields'] = False
            template_data['models'][model_key]['has_foreign_keys'] = False
            template_data['models'][model_key]['fields'] = list()

            for field in app_data['models'][model_key]['fields']:
                field_dict = dict()
                field_dict['field_name'] = field['field_name']
                field_dict['type'] = field['type']
                field_dict['is_included'] = True
                if field.get('remote_field'):
                    field_dict['foreign_key'] = field.get('remote_field')
                    template_data['models'][model_key]['has_foreign_keys'] = True
                if field['type'] in self.field_types_to_ignore:
                    field_dict['is_included'] = False
                    field_dict['is_ignored'] = True
                    template_data['models'][model_key]['has_ignored_fields'] = True

                template_data['models'][model_key]['fields'].append(field_dict)

                # else:
                #     method_name = 'get_{}_factory'.format(field['type'].lower())
                #     if hasattr(self, method_name):
                #         field_dict['factory'] = getattr(self, method_name)(field, template_data['models'][model_key])
                #         template_data['models'][model_key]['fields'].append(field_dict)
                #     else:
                #         field_dict['is_supported'] = False
                #         template_data['models'][model_key]['fields'].append(field_dict)
        return template_data



import os

from django.conf import settings
from django.core.management import BaseCommand

from django_test_tools.generators.crud_generator import GenericTemplateWriter
from django_test_tools.generators.model_generator import FactoryBoyGenerator
from ...app_manager import DjangoAppManager

PRINT_IMPORTS = """
import string

from random import randint
from pytz import timezone

from django.conf import settings

from factory import Iterator
from factory import LazyAttribute
from factory import SubFactory
from factory import lazy_attribute
from factory.django import DjangoModelFactory, FileField
from factory.fuzzy import FuzzyText, FuzzyInteger
from faker import Factory as FakerFactory

faker = FakerFactory.create()

"""
PRINT_FACTORY_CLASS = """
class {0}Factory(DjangoModelFactory):
    class Meta:
        model = {0}
"""

PRINT_CHARFIELD = """    {} = LazyAttribute(lambda x: faker.text(max_nb_chars={}))"""
PRINT_CHARFIELD_NUM = """    {} = LazyAttribute(lambda x: FuzzyText(length={}, chars=string.digits).fuzz())"""
PRINT_CHARFIELD_LETTERS = """    {} = LazyAttribute(lambda x: FuzzyText(length={}, chars=string.ascii_letters).fuzz())"""
PRINT_CHARFIELD_CHOICES = """    {} = Iterator({}.{}, getter=lambda x: x[0])"""
PRINT_DATETIMEFIELD = """    {} = LazyAttribute(lambda x: faker.date_time_between(start_date="-1y", end_date="now",
                                                           tzinfo=timezone(settings.TIME_ZONE)))"""
PRINT_FOREIGNKEY = """    {} = SubFactory({}Factory){}"""
PRINT_FILEFIELD = """    {} = FileField(filename='{}.{}')"""
PRINT_BOOLEANFIELD = """    {} = Iterator([True, False])"""
PRINT_INTEGERFIELD = """    {} = LazyAttribute(lambda o: randint(1, 100))"""
PRINT_IPADDRESSFIELD = """    {} = LazyAttribute(lambda o: faker.ipv4(network=False))"""
PRINT_TEXTFIELD = """    {} = LazyAttribute(lambda x: faker.paragraph(nb_sentences=3, variable_nb_sentences=True))"""
PRINT_DECIMALFIELD = """    {} = LazyAttribute(lambda x: faker.pydecimal(left_digits={}, right_digits={}, positive=True))"""
PRINT_UNSUPPORTED_FIELD = """    #{} = {} We do not support this field type"""
PRINT_COUNTRYFIELD = """    {} = Iterator(['PA', 'US'])"""


# noinspection PyProtectedMember
class ModelFactoryGenerator(object):
    def __init__(self, model):
        self.model = model

    def _generate(self):
        factory_class_content = list()
        factory_class_content.append({'print': PRINT_FACTORY_CLASS, 'args': [self.model.__name__]})
        for field in self.model._meta.fields:
            field_type = type(field).__name__
            field_data = dict()
            if field_type in ['AutoField', 'AutoCreatedField', 'AutoLastModifiedField']:
                pass
            elif field_type in ['DateTimeField', 'DateField']:
                field_data = {'print': PRINT_DATETIMEFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)
            elif field_type == 'CharField':
                field_data = self._get_charfield(field)
                factory_class_content.append(field_data)
            elif field_type == 'ForeignKey':
                related_model = field.remote_field.model.__name__
                field_data = {'print': PRINT_FOREIGNKEY,
                              'args': [field.name, related_model, '']}
                factory_class_content.append(field_data)
            elif field_type == 'BooleanField':
                field_data = {'print': PRINT_BOOLEANFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)

            elif field_type == 'TextField':
                field_data = {'print': PRINT_TEXTFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)

            elif field_type == 'IntegerField':
                field_data = {'print': PRINT_INTEGERFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)

            elif field_type == 'FileField':
                field_data = {'print': PRINT_FILEFIELD, 'args': [field.name, field.name, 'xlsx']}
                factory_class_content.append(field_data)

            elif field_type == 'DecimalField' or field_type == 'MoneyField':
                max_left = field.max_digits - field.decimal_places
                max_right = field.decimal_places
                field_data = {'print': PRINT_DECIMALFIELD,
                              'args': [field.name, max_left, max_right]}
                factory_class_content.append(field_data)

            elif field_type == 'GenericIPAddressField':
                field_data = {'print': PRINT_IPADDRESSFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)

            elif field_type == 'CountryField':
                field_data = {'print': PRINT_COUNTRYFIELD, 'args': [field.name]}
                factory_class_content.append(field_data)
            else:
                field_data = {'print': PRINT_UNSUPPORTED_FIELD,
                              'args': [field.name, field_type]}
                factory_class_content.append(field_data)

        return factory_class_content

    def _get_charfield(self, field):
        field_data = dict()

        if field.choices is not None and len(field.choices) > 0:
            field_data = {'print': PRINT_CHARFIELD_CHOICES, 'args': [field.name, self.model.__name__, 'CHOICES']}
            return field_data
        else:
            if self._is_number(field.name):
                field_data = {'print': PRINT_CHARFIELD_NUM,
                              'args': [field.name, field.max_length]}
                return field_data
            else:
                if field.max_length >= 5:
                    field_data = {'print': PRINT_CHARFIELD,
                                  'args': [field.name, field.max_length]}
                else:
                    field_data = {'print': PRINT_CHARFIELD_LETTERS,
                                  'args': [field.name, field.max_length]}
                return field_data

    def _is_number(self, field_name):
        num_vals = ['id', 'num']
        for nv in num_vals:
            if nv in field_name.lower():
                return True
        return False

    def __str__(self):
        printable = list()
        for print_data in self._generate():
            try:
                printable.append(print_data['print'].format(*print_data['args']))
            except IndexError as e:
                print('-' * 74)
                print('{print} {args}'.format(**print_data))
                raise e

        return '\n'.join(printable)


class Command(BaseCommand):
    """

        $ python manage.py generate_factories project.app
    """

    def add_arguments(self, parser):
        parser.add_argument('app_name')
        # parser.add_argument("-l", "--list",
        #                     action='store_true',
        #                     dest="list",
        #                     help="List employees",
        #                     )
        # parser.add_argument("-a", "--assign",
        #                     action='store_true',
        #                     dest="assign",
        #                     help="Create unit assignments",
        #                     )
        #
        #
        parser.add_argument("--filename",
                            dest="filename",
                            help="Output filename",
                            default=None,
                            )
        # parser.add_argument("--start-date",
        #                     dest="start_date",
        #                     help="Start date for the assignment",
        #                     default=None,
        #                     )
        # parser.add_argument("--fiscal-year",
        #                     dest="fiscal_year",
        #                     help="Fiscal year for assignments",
        #                     default=None,
        #                     )
        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        app_name = options.get('app_name')
        if options.get('filename'):
            filename = os.path.join(settings.TEST_OUTPUT_PATH, options.get('filename'))
            generator = FactoryBoyGenerator()
            factory_data = generator.create_template_data(app_name)
            template_name = 'factories.py.j2'
            writer = GenericTemplateWriter(template_name)
            writer.write(factory_data, filename)
        else:
            app_manager = DjangoAppManager()
            app = app_manager.get_app(app_name)
            if not app:
                self.stderr.write('This command requires an existing app name as '
                                  'argument')
                self.stderr.write('Available apps:')
                for app in sorted(app_manager.installed_apps):
                    self.stderr.write('    %s' % app)
            else:
                self.stdout.write(PRINT_IMPORTS)
                for model in app.get_models():
                    model_fact = ModelFactoryGenerator(model)
                    self.stdout.write(str(model_fact))



from django.core.management import BaseCommand

from ...file_utils import shorten_path
from ...pip.utils import update_outdated_libraries


class Command(BaseCommand):
    """
        $ python manage.py
    """

    def add_arguments(self, parser):
        parser.add_argument('requirement_filename')


    def handle(self, *args, **options):
        changes = update_outdated_libraries(options['requirement_filename'])
        for change in changes:
            change['filename'] = shorten_path(change['filename'])
            self.stdout.write('Changed {library_name} in file {filename} to {new}'.format(**change))








import hashlib
import json
import os
import pickle
import shutil
from datetime import date, datetime

from django.conf import settings
from django.utils import timezone

from django_test_tools.exceptions import DjangoTestToolsException

BLOCKSIZE = 65536


def create_dated(filename):
    """
    Based on the filename will create a full path filename including the date and time in '%Y%m%d_%H%M' format.
    The path to the filename will be set in the TEST_OUTPUT_PATH settings variable.

    If the TEST_OUTPUT_PATH folder doesn't exist the function will create it.

    :param filename: base filename. my_excel_data.xlsx for example
    :return: string, full path to file with date and time in the TEST_OUTPUT_PATH folder
    """
    if getattr(settings, 'TEST_OUTPUT_PATH', None) is None:
        msg = 'You need a the variable TEST_OUTPUT_PATH in settings. It should point to a folder' \
              'for temporary data to be written and reviewed.'
        raise ValueError(msg)
    if not os.path.exists(settings.TEST_OUTPUT_PATH):
        os.makedirs(settings.TEST_OUTPUT_PATH)
    return add_date(os.path.join(settings.TEST_OUTPUT_PATH, filename))


def hash_file(filename, algorithm='sha1', block_size=BLOCKSIZE):
    """
    Creates a unique hash for a file.

    :param filename: String with the full path to the file
    :param algorithm: String Algorithm to create the hash
    :param block_size: int for the size of the block while reading the file
    :return: string the hash for the file
    """
    try:
        hasher = getattr(hashlib, algorithm)()
    except AttributeError:
        raise ValueError('{} is not a valid hashing algorithm'.format(algorithm))

    with open(filename, 'rb') as afile:
        buf = afile.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(block_size)
    return hasher.hexdigest()


def parametrized(dec):
    """
    Need to study this code.
    Got it from http://stackoverflow.com/questions/5929107/python-decorators-with-parameters

    :param dec:
    :return:
    """

    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)

        return repl

    return layer


@parametrized
def temporary_file(func, extension, delete_on_exit=True):
    """
    This method decorator creates a filename with date using the provided extension and delete the file after the method
    has been executed.

    The settings.TEST_OUTPUT_PATH must be configured in your settings file.

    .. code-block:: python

        @temporary_file('json')
        def test_temporary_file_decorator(self):
            filename = self.test_temporary_file_decorator.filename
            ... write to the file ...

    :param func: function to decorate
    :param extension: extension of the filename without the dot
    :param delete_on_exit: If True the filename will be deleted.
    :return: the function
    """
    filename = create_dated('{}.{}'.format(func.__name__, extension))

    def function_t_return(*args):
        results = func(*args)
        if os.path.exists(filename) and delete_on_exit:
            os.remove(filename)
        return results

    function_t_return.filename = filename
    return function_t_return

@parametrized
def temporary_files(func, extension, delete_on_exit=True, count=2):
    """
    This method decorator creates a filename with date using the provided extension and delete the file after the method
    has been executed.

    The settings.TEST_OUTPUT_PATH must be configured in your settings file.

    .. code-block:: python

        @temporary_files('json')
        def test_temporary_file_decorator(self):
            filename = self.test_temporary_file_decorator.filenames[0]
            ... write to the file ...

    :param func: function to decorate
    :param extension: extension of the filename without the dot
    :param delete_on_exit: If True the filename will be deleted.
    :return: the function
    """
    filenames = list()
    for i in range(count):
        filename = create_dated('{}-{}.{}'.format(func.__name__, i, extension))
        filenames.append(filename)

    def function_t_return(*args):
        results = func(*args)
        for filename in filenames:
            if os.path.exists(filename) and delete_on_exit:
                os.remove(filename)
            return results

    function_t_return.filenames = filenames
    return function_t_return
def shorten_path(path, level=2, current_level=1):
    """
    This method shortens the path by eliminating the folders on top.

    .. code-block:: python

        filename = '/user/documents/personal/file.txt'
        shortened = shorten_path(filename)
        self.assertEqual(shortened, 'personal/file.txt')


    :param path: string full path for the filename
    :param level: int, number of levels to show.
    :param current_level: int, recursing level.
    :return: string shortened path
    """
    if level == 0:
        raise ValueError('The minimum level accepted is one')
    path, tail = os.path.split(path)
    if level == current_level:
        return tail
    else:
        if path != os.path.sep:
            return shorten_path(path, level, current_level + 1) + os.path.sep + tail
        return tail


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code
    taken from: https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable
    """

    if isinstance(obj, (datetime, date)):
        serial = obj.isoformat()
        return serial
    raise TypeError("Type %s not serializable" % type(obj))


def serialize_data(data, output_file=None, format='json', **kwargs):
    """
    Quick function to serialize a data to file. The data keys will be saved in an alphabetical order
    for consistency purposes.
    If no output_file is supplied the function will created a dated file in the settings.TEST_OUTPUT_PATH folder.
    if the output_file is a folder the dated file will be created on the supplied folder with the serialized date.
    if the output_file is a file the data will be serialized to thar file

    :param data: Dictionary or list to serialize
    :param format: Format to serialize to. Currently json is the only one supported
    :param output_file: File to output the data to
    :param kwargs:
    """
    assert format in ['json', 'pickle'], 'Unsupported format {}'.format(format)
    base_filename = kwargs.get('base_filename', 'serialized_data')

    if output_file is None:
        filename = create_dated('{}.{}'.format(base_filename, format))
    elif os.path.isdir(output_file):
        filename = os.path.join(output_file, '{}.{}'.format(base_filename, format))
    else:
        filename = output_file
    if format == 'json':
        with open(filename, 'w', encoding=kwargs.get('encoding', 'utf-8'), newline='\n') as fp:
            json.dump(data, fp, indent=kwargs.get('indent', 4),
                      default=json_serial, sort_keys=True)
    elif format == 'pickle':
        with open(filename, 'wb') as output:
            pickle.dump(data, output, pickle.HIGHEST_PROTOCOL)

    return filename


def add_date(filename, **kwargs):
    """
    Adds to a filename the current date and time in '%Y%m%d_%H%M' format.
    For a filename /my/path/myexcel.xlsx the function would return /my/path/myexcel_20170101_1305.xlsx.
    If the file already exists the function will add seconds to the date to attempt to get a unique name.

    The function will detect if another file exists with the same name if it exist it will append seconds to the
    filename. For example if file /my/path/myexcel_20170101_1305.xlsx alread exist thte function will return
    /my/path/myexcel_20170101_130521.xlsx.

    :param filename: string with fullpath to file or just the filename
    :param kwargs: dictionary. date_position: suffix or preffix, extension: string to replace extension
    :return: string with full path string including the date and time
    """
    current_datetime = timezone.localtime(timezone.now()).strftime('%Y%m%d_%H%M%S')
    new_filename_data = dict()
    suffix_template = '{path}{separator}{filename_with_out_extension}_{datetime}.{extension}'
    prefix_template = '{path}{separator}{datetime}_{filename_with_out_extension}.{extension}'
    if '/' in filename and '\\' in filename:
        raise ValueError('Filename %s contains both / and \\ separators' % filename)
    if '\\' in filename:
        path_parts = filename.split('\\')
        file = path_parts[-1]
        path = '\\'.join(path_parts[:-1])
        separator = '\\'
    elif '/' in filename:
        path_parts = filename.split('/')
        file = path_parts[-1]
        path = '/'.join(path_parts[:-1])
        separator = '/'
    else:
        file = filename
        path = ''
        separator = ''

    new_filename_data['path'] = path
    parts = file.split('.')
    if kwargs.get('extension', None) is not None:
        new_filename_data['extension'] = kwargs['extension']
    else:
        if len(parts) > 1:
            new_filename_data['extension'] = parts[-1]
        else:
            new_filename_data['extension'] = ''

    new_filename_data['separator'] = separator
    if new_filename_data['extension'] == '':
        new_filename_data['filename_with_out_extension'] = parts[0]
    else:
        new_filename_data['filename_with_out_extension'] = '.'.join(parts[:-1])
    new_filename_data['datetime'] = current_datetime[:-2]  # Seconds are stripped

    date_position = kwargs.get('date_position', 'suffix')
    if date_position == 'suffix':
        new_filename = suffix_template.format(**new_filename_data)
        if os.path.exists(new_filename):
            new_filename_data['datetime'] = current_datetime
            new_filename = suffix_template.format(**new_filename_data)
        if new_filename_data['extension'] == '':
            new_filename = new_filename[:-1]
    else:
        new_filename = prefix_template.format(**new_filename_data)
        if os.path.exists(new_filename):
            new_filename_data['datetime'] = current_datetime
            new_filename = prefix_template.format(**new_filename_data)
        if new_filename_data['extension'] == '':
            new_filename = new_filename[:-1]

    return new_filename


class TemporaryFolder:
    def __init__(self, base_name, delete_on_exit=True):
        self.new_path = create_dated(base_name)
        self.delete_on_exit = delete_on_exit

    def __enter__(self):
        os.mkdir(self.new_path)
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.saved_path)
        if self.delete_on_exit:
            shutil.rmtree(self.new_path)

    def write(self, filename, content):
        with open(filename, 'w', encoding='utf-8') as file:
            if isinstance(content, str):
                file.writelines(content)
            elif isinstance(content, list):
                for line in content:
                    file.write(line)
                    file.write('\n')
            else:
                file.writelines(str(content))
        return os.path.join(self.new_path, filename)


def compare_file_content(*args, **kwargs):
    errors = list()
    file1 = args[0]
    file2 = args[1]
    excluded_lines = kwargs.get('excluded_lines', [])
    encoding = kwargs.get('encoding', 'utf-8')
    raise_exception = kwargs.get('raise_exception', True)
    eol = kwargs.get('eol', '\n')

    def get_lines(filename):
        with open(filename, 'r', encoding=encoding, newline=eol) as file:
            lines = file.readlines()
        return lines

    lines1 = get_lines(file1)
    lines2 = get_lines(file2)

    for i in range(len(lines1)):
        if i not in excluded_lines:
            if lines1[i] != lines2[i]:
                msg = 'On line {} expected "{}" got "{}"'.format(i,
                                                                 lines1[i].replace(eol, ''),
                                                                 lines2[i].replace(eol, ''))
                errors.append(msg)
                if raise_exception:
                    raise DjangoTestToolsException(msg)
    return errors



import pytz
from datetime import datetime
import os

import subprocess
from django.conf import settings
from django.core.management import BaseCommand

import logging
from openpyxl import Workbook

from ...file_utils import add_date
from ...utils import datetime_to_local_time

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    This command will export your git commits into a simple excel sheet. The report includes hash, email, date and
    description.

    Instead of using:

        git log --date=iso --pretty="format:%h|%ae|%ai|%s" > ./output/commits.txt

    You can type

        $ python manage.py git_report -f /path/to/excel.xlsx

    Note: usefulll datetime format in excel dd-mmm-yy h:mm:ss

    """

    def add_arguments(self, parser):
        # parser.add_argument('requirement_filename')
        # parser.add_argument("-l", "--list",
        #                     action='store_true',
        #                     dest="list",
        #                     help="List data",
        #                     )

        parser.add_argument("--format",
                            dest="format",
                            help="Git format",
                            default=None,
                            )

        parser.add_argument('-f', "--filename",
                            dest="filename",
                            help="Output filename",
                            default=None,
                            )

        # parser.add_argument("-u", "--username",
        #                 dest="usernames",
        #                 help="LDAP usernames for employees",
        #                 nargs='+',
        #                 )

    def handle(self, *args, **options):
        headers = ['Hash', 'Email', 'Date', 'Description']
        if options.get('format') is None:
            git_format = '%h|%ae|%ai|%s'
        else:
            git_format = options.get('format')

        if options.get('filename') is None:
            filepath = os.path.join(settings.TEST_OUTPUT_PATH, 'git_report.xlsx')
            filename = add_date(filepath)
        else:
            filename = options.get('filename')

        wb = Workbook()
        sheet = wb.create_sheet()
        row = 1
        col = 1
        for header in headers:
            sheet.cell(row=row, column=col, value=header)
            col += 1
        row += 1
        git_lines = self.report(git_format)
        for line in git_lines:
            self.stdout.write(line)
            line_data = line.split('|')
            col = 1
            for data in line_data:
                if col == 3:
                    data = self.parse_date(data)
                sheet.cell(row=row, column=col, value=data)
                col += 1
            row += 1

        wb.save(filename)

        self.stdout.write('Finished processing {}'.format(filename))

    def report(self, git_format):
        # git log --pretty=format:"%h - %an, %ad : %s" --date=iso
        #
        # git log --pretty="%h - %s" --author=gitster --since="2008-10-01"
        # --before="2008-11-01" --no-merges -- t/
        try:
            # git-describe doesn't update the git-index, so we do that
            # subprocess.check_output(["git", "update-index", "--refresh"])

            # get info about the latest tag in git
            git_commands = ["git", "log", '--date=iso', '--pretty=format:"{}"'.format(git_format)]

            describe_out = subprocess.check_output(
                git_commands,
            ).decode('utf-8')
        except subprocess.CalledProcessError:
            logger.warning("Error when running git describe")
            return {}

        return self.cleanup(describe_out.split('\n'))

    def cleanup(self, git_lines):
        cleaned_up = list()
        for git_line in git_lines:
            cleaned_up.append(git_line[1:-1])
        return cleaned_up


    def parse_date(self, date_value):
        """
        Will transform a string in the format '2018-09-23 09:37:50 -0500' to a datetime object
        :param date_value: <str> date in iso format
        :return: <datetime> datetime object of the string value
        """
        date_format = '%Y-%m-%d %H:%M:%S %z'

        datetime_object = datetime.strptime(date_value, date_format)
        datetime_object = datetime_object.replace(tzinfo=pytz.UTC)
        time_zone = pytz.timezone(settings.TIME_ZONE)

        datetime_object_with_timezone = datetime_object.astimezone(time_zone)
        return datetime_object



from openpyxl import load_workbook


class ExcelAdapter(object):

    @classmethod
    def convert_to_list(cls, filename, sheet_name=None, has_header=True):
        """
        Reads an Excel file and converts every row into a dictionary. All values are converted to strings.
        Assumes first row contains the name of the attributes.

        :param filename: <str> Excel filename
        :param sheet_name: <str> Name of the sheet
        :return: <list> A list of dictionaries.
        """
        data = list()
        wb = load_workbook(filename=filename, data_only=True)
        if sheet_name is None:
            sheet = wb.active
        else:
            sheet = wb[sheet_name]
        row_num = 1
        for row in sheet.rows:
            row_data = list()
            if row_num == 1 and has_header:
                row_num += 1
                continue
            for column in range(0, len(row)):
                row_data.append(row[column].value)
            data.append(row_data)
            row_num += 1
        return data

    @classmethod
    def convert_to_dict(cls, filename, sheet_name=None):
        """
        Reads an Excel file and converts every row into a dictionary. All values are converted to strings.
        Assumes first row contains the name of the attributes.

        :param filename: <str> Excel filename
        :param sheet_name: <str> Name of the sheet
        :return: <list> A list of dictionaries.
        """
        data = list()
        wb = load_workbook(filename=filename, data_only=True)
        if sheet_name is None:
            sheet = wb.active
        else:
            sheet = wb[sheet_name]
        row_num = 1
        attributes = list()
        for row in sheet.rows:
            row_data = dict()
            if row_num == 1:
                row_num += 1
                for column in range(0, len(row)):
                    attributes.append(row[column].value)
                continue
            for column in range(0, len(row)):
                row_data[attributes[column]] = str(row[column].value)
            data.append(row_data)
            row_num += 1
        return data



import importlib

from django.core.management import BaseCommand

from ...generators.crud_generator import SerializerTestGenerator


class Command(BaseCommand):
    """

        $ python manage.py generate_serializers_test my_application.app.api.serializers.MyObjectSerializer -f output/m.py
    """

    def add_arguments(self, parser):
        parser.add_argument('serializer_class')

        parser.add_argument("-f", "--filename",
                            dest="filename",
                            help="Output filename",
                            )

    def handle(self, *args, **options):
        parts = options.get('serializer_class').split('.')
        serializer_class_name = parts[-1]
        module_name = '.'.join(parts[0:-1])
        filename = options.get('filename')
        generator = SerializerTestGenerator()
        data = dict()
        my_module = importlib.import_module(module_name)
        MySerializer = getattr(my_module, serializer_class_name)

        serializer = MySerializer()
        rep_ser = repr(serializer)
        fields = list()
        str_fields = list()
        for field in serializer.fields:
            fields.append(field)

        for key, field in serializer.fields.fields.items():
            if type(field).__name__ in ['CharField', 'TextField']:
                str_fields.append(field.field_name)
            print(key)

        data['model_name'] = serializer.Meta.model.__name__
        data['fields'] = fields
        data['string_vars'] = str_fields
        generator.print(data, filename)
        self.stdout.write('Printed to file {}'.format(filename))
